from typing import (
    List,
    Dict,
    Tuple,
    Set,
    Optional
)

from .util import (
    apply_precision
)

from .types import (
    Balance,
    SymbolPosition,
    OrderTicket,
    TicketOrderSide,
    TicketOrderStatus,

    TicketGroup,
    SymbolInfo,

    FuncCancelOrder,

    FLOAT_ZERO
)


CreateTicketReturn = Tuple[
    Optional[OrderTicket],
    Set[OrderTicket]
]


"""
Terminology
- asset: BTC, USDT, or others
- symbol: trading pairs of assets

Principle
- be pure
- be sync
- be passive, no triggers
- no default parameters for public methods
"""


class TraderState:
    """State Phase II

    - only support base asset position 0 or 1
    - support multiple base assets
    - only support a single quote asset

    Convention:
    - For a certain base asset,
      its related tickets should has the same direction

    Design principle:
    - The expectation settings are the final state that the system try to achieve.
    """

    _balances: Dict[str, Balance]

    _expected: Dict[str, SymbolPosition]
    _old_expected: Dict[str, SymbolPosition]

    _tickets: Dict[
        # asset
        str,
        TicketGroup
    ]

    _symbol_prices: Dict[str, float]
    _symbol_infos: Dict[str, SymbolInfo]

    _cancel_order: Optional[FuncCancelOrder]
    _all_tickets: Set[OrderTicket]

    def __init__(self) -> None:
        # Asset quota dict: asset -> quantity
        self._quotas = {}

        self._balances = {}
        self._expected = {}

        # In the beginning, they are the same
        self._old_expected = self._expected

        self._symbol_prices = {}

        self._symbol_infos = {}

        self._tickets = {}
        self._all_tickets = set()

    # Public methods
    # ------------------------------------------------------------------------

    def set_price(
        self,
        symbol_name: str,
        price: float
    ) -> bool:
        """
        Set the price of a symbol

        Returns `True` if the price changes
        """

        old_price = self._symbol_prices.get(symbol_name)

        if price == old_price:
            # If the price does not change, should not reset diff
            return False

        self._symbol_prices[symbol_name] = price
        self._reset_diff()

        return True

    def set_symbol_info(
        self,

        # symbol name, ie the trading pair name, e.g. BTCUSDT
        symbol_name: str,

        # base asset, ie the coin, token or asset currency to trade, e.g. BTC
        base_asset: str,

        # quote asset, ie the asset to quote, quote currency, e.g. USDT
        quote_asset: str,

        # Price filters
        # ------------------------------------------------
        min_price: str,
        max_price: str,

        # min ticker size
        min_price_step: str,

        # Lot size filters
        # ------------------------------------------------
        min_quantity: str,
        max_quantity: str,
        min_quantity_step: str,

        min_notional: str
    ) -> None:
        """
        Set trading info of a symbol

        Usage::

        state.set_symbol_info(
            'BTCUSDT',
            'BTC',
            'USDT',

            '0.01',
            '1000000',
            '0.01',

            '0.000001',
            '9000',
            '0.000001',

            '10.00000000'
        )
        """

        self._symbol_infos[symbol_name] = SymbolInfo(
            symbol_name,
            base_asset,
            quote_asset,

            min_price,
            max_price,
            min_price_step,

            min_quantity,
            max_quantity,
            min_quantity_step,

            min_notional
        )

    def set_quota(
        self,
        asset: str,
        quota: Optional[float]
    ) -> None:
        """
        Set the quota for a certain asset.

        The quota of an asset limits
        the maximum quantity the trader could **BUY** the asset,
        but DON'T limit sell.

        Args:
            quota (float): how much available base quote asset buy the asset

        For example, if

        - quota of BTC: $35000
        - current price: 7000
        - cash quantity: $70000

        Then, the trader could only buy 5 BTC,
        although the cash is enough to buy 10 BTC
        """

        old_quota = self._quotas.get(asset)

        self._quotas[asset] = quota

        if old_quota != quota:
            self._reset_diff()

    def update_balances(
        self,
        new: List[Balance]
    ) -> None:
        """
        Update user balances, including normal assets and cash asset(USDT)

        Usage::

            state.update([
                ('BTC', 8.)
            ])
        """

        balances = []

        for balance in new:
            balances.append(
                (balance.asset, balance)
            )

        self._balances.update(balances)

        # This will trigger diffing again
        self._reset_diff()

    def support_symbol(self, symbol_name: str) -> bool:
        """
        Check whether the symbol is supported
        """

        return symbol_name in self._symbol_infos

    def expect(
        self,
        symbol_name: str,
        position: float,
        asap: bool,
        price: Optional[float]
    ) -> bool:
        """
        Update expectation, returns whether it is successfully updated

        Usage::

            # to all-in BTC
            state.expect('BTCUSDT', 1.)

        Args:
            price (:obj:`float`, optional): the price to expect. For now, this argument is not actually used because the current strategy is not based on price prediction but technical shape.
        """

        symbol_info = self._get_symbol_info(symbol_name)

        if symbol_info is None:
            # This is not duplicated with `self.support_symbol()`
            # TraderState not cares about the usage of business, so it provides
            #   several util methods
            # The same is true for `self.create_ticket()`
            return False

        # Create a new dict so that will be considered as changed
        self._expected = {
            **self._expected
        }

        self._expected[symbol_name] = SymbolPosition(
            symbol_info.symbol,
            position,
            asap,
            price
        )

        return True

    def clear(self) -> None:
        """
        Clear all expectations and tickets
        """

        self._expected.clear()
        self._all_tickets.clear()

        for group in self._tickets.values():
            group.clear()

    def create_ticket(
        self,
        symbol_name: str,
        position: float,
        asap: bool,
        price: Optional[float]
    ) -> CreateTicketReturn:
        """
        Create a ticket according to the position of a symbol, and remove position expectations of the symbol

        Returns:
            tuple: An optional newly-created ticket and tickets which needs to be cancelled
        """

        symbol_info = self._get_symbol_info(symbol_name)

        if symbol_info is None:
            # Symbol is not supported
            return None, set()

        # Remove symbol position expectation
        self._expected.pop(symbol_name, None)

        position = SymbolPosition(
            symbol_info.symbol,
            position,
            asap,
            price
        )

        return self._create_ticket(position)

    def get_tickets(
        self,
        status: TicketOrderStatus = TicketOrderStatus.INIT
    ) -> Tuple[
        List[OrderTicket],
        Set[OrderTicket]
    ]:
        """
        Get all unsubmitted tickets which are not submitted

        Args:
            status (:obj:`TicketOrderStatus`, optional): The maximun status phase that the returned tickets will have. Defaults to `TicketOrderStatus.INIT`

        Returns
            tuple: a list of available tickets and a set of tickets to cancel
        """

        tickets_to_cancel = self._diff()

        return [
            ticket
            for ticket in self._all_tickets

            # Only collect tickets whose status are <= `status`
            if ticket.status.lt(status)
        ], tickets_to_cancel

    def close_ticket(
        self,
        ticket: OrderTicket
    ) -> None:
        """
        Just close a ticket

        Args:
            ticket (OrderTicket): the ticket to close
            position_fulfilled (:obj:`bool`, optional):
        """

        self._remove_expectation(ticket.position)

        self._get_ticket_group(ticket.symbol.base_asset).close(ticket)
        self._get_ticket_group(ticket.locked_asset).close(ticket)
        self._all_tickets.discard(ticket)

    # End of public methods ---------------------------------------------

    def _remove_expectation(self, position: SymbolPosition) -> None:
        symbol_name = position.symbol.name

        existed = self._expected.get(symbol_name)

        if existed is not None and existed.equals_to(position):
            del self._expected[symbol_name]

    def _reset_diff(self):
        self._old_expected = {}

    def _get_symbol_info(self, symbol_name: str) -> Optional[SymbolInfo]:
        return self._symbol_infos.get(symbol_name)

    def _get_ticket_group(self, asset: str) -> TicketGroup:
        group = self._tickets.get(asset)

        if group is None:
            group = TicketGroup()
            self._tickets[asset] = group

        return group

    def _get_ticket_locked(self, asset: str) -> float:
        group = self._get_ticket_group(asset)

        locked = FLOAT_ZERO

        if not group.sell:
            return locked

        for ticket in group.sell:
            locked += ticket.locked_quantity

        return locked

    def _get_balance(self, asset: str) -> Tuple[float, float]:
        """
        The balance updates are not always in time, so that we should double confirm the locked quantity of a asset.
        This method will return the ensured tuple of free and locked quantity
        """

        balance = self._balances.get(asset)

        if balance is None:
            # We dont have this asset
            return FLOAT_ZERO, FLOAT_ZERO

        locked_by_tickets = self._get_ticket_locked(asset)

        delta = balance.locked - locked_by_tickets

        if delta >= FLOAT_ZERO:
            # The locked quantity has been updated
            return balance.free, balance.locked

        # The locked quantity has not been updated,
        # we use the larget locked quantity
        return balance.free + delta, locked_by_tickets

    def _adjust_quote_quantity(
        self,
        asset: str,
        price: float,
        quantity: float
    ) -> float:
        """
        Adjust the quote quantity according to quota
        """

        quota = self._quotas.get(asset)

        if quota is None:
            return quantity

        free, locked = self._get_balance(asset)

        return max(
            min(quantity, quota) - (free + locked) * price,
            0.
        )

    def _diff(self) -> Set[OrderTicket]:
        """
        Diff the position expectations based on what we create tickets

        Returns
            set: a set of tickets to cancel
        """

        tickets_to_cancel = set()

        if self._expected is self._old_expected:
            # If there is no new expectation,
            # so skip diffing
            return tickets_to_cancel

        self._old_expected = self._expected

        if self._expected:
            for position in self._expected.values():
                _, to_cancel = self._create_ticket(position)
                tickets_to_cancel |= to_cancel

        return tickets_to_cancel

    def _create_ticket(
        self,
        position: SymbolPosition
    ) -> CreateTicketReturn:
        """
        Create a ticket from a symbol position
        """

        symbol = position.symbol
        symbol_name = symbol.name
        asset = symbol.base_asset
        group = self._get_ticket_group(asset)

        tickets_to_cancel = self._get_tickets_to_cancel(
            group,
            position,
            TicketOrderSide.SELL if position.value == FLOAT_ZERO
            else TicketOrderSide.BUY
        )

        # `info` is always not None, which is ensured by self.expect()
        info = self._get_symbol_info(symbol_name)

        price = (
            # Order with specified price
            position.price
            or self._symbol_prices.get(symbol_name)
        )

        if price is None:
            # The price is not ready,
            # then skip to avoid unexpected result
            return None, tickets_to_cancel

        if position.value == FLOAT_ZERO:
            # Sell all out
            # ---------------------------------------------------------------

            quantity, _ = self._get_balance(asset)

            if quantity < info.min_quantity_step:
                # is less than minimum trading quantity
                return None, tickets_to_cancel

            quantity_str = apply_precision(
                quantity,
                info.min_quantity_step_precision
            )

            # Apply precision
            quantity = float(quantity_str)

            ticket = OrderTicket(
                TicketOrderSide.SELL,
                symbol,
                price,
                quantity_str,
                asset,
                quantity,
                position,
                info  # type: ignore
            )

            self._get_ticket_group(asset).sell.add(ticket)

        else:
            # Buy all in
            # ---------------------------------------------------------------

            quote_asset = symbol.quote_asset

            free, _ = self._get_balance(quote_asset)

            available = self._adjust_quote_quantity(asset, price, free)

            if available < info.min_price_step:
                # The quantity of the current cash is less than
                # the minimun trading price
                return None, tickets_to_cancel

            if available < info.min_notional:
                return None, tickets_to_cancel

            if free < price * info.min_quantity_step:
                # The quantity of the current cash is less than
                # the value minium asset quantity
                return None, tickets_to_cancel

            quantity_str = apply_precision(
                available / price,
                info.min_quantity_step_precision
            )

            locked_quote_quantity = price * float(quantity_str)

            ticket = OrderTicket(
                TicketOrderSide.BUY,
                symbol,
                price,
                quantity_str,
                quote_asset,
                locked_quote_quantity,
                position,
                info  # type: ignore
            )

            self._get_ticket_group(asset).buy.add(ticket)
            self._get_ticket_group(quote_asset).sell.add(ticket)

        self._all_tickets.add(ticket)

        return ticket, tickets_to_cancel

    def _get_tickets_to_cancel(
        self,
        group: TicketGroup,
        position: SymbolPosition,
        direction: TicketOrderSide
    ) -> Set[OrderTicket]:
        opposite_group = group.get(
            TicketOrderSide.SELL if direction is TicketOrderSide.BUY
            else TicketOrderSide.BUY
        )

        if opposite_group:
            # If tickets of opposite direction exist,
            # then cancel them
            for ticket in opposite_group:
                self.close_ticket(ticket)

            return opposite_group

        to_cancel = set()
        tickets = group.get(direction)

        if position.price is not None:
            for ticket in tickets:
                if not ticket.position.equals_to(position):
                    to_cancel.add(ticket)
                    self.close_ticket(ticket)

        return to_cancel



from typing import (
    Set,
    Callable,
    Awaitable,
    Optional
)

from enum import Enum

from .util import (
    class_repr,
    float_to_str,
    # datetime_now_str
)

FLOAT_ZERO = 0.


class Symbol:
    __slots__ = (
        'name',
        'base_asset',
        'quote_asset'
    )

    name: str
    base_asset: str
    quote_asset: str

    def __init__(
        self,
        name: str,
        base_asset: str,
        quote_asset: str
    ) -> None:
        self.name = name
        self.base_asset = base_asset
        self.quote_asset = quote_asset

    def __repr__(self) -> str:
        return self.name


class SymbolPosition:
    """The position ratio relative to the whole balance value

    Args:
        value (float): should between 0 and 1, for now only support 0 or 1
    """

    __slots__ = (
        'symbol',
        'value',
        'asap',
        'price'
    )

    symbol: Symbol
    value: float
    asap: bool
    price: Optional[float]

    def __init__(
        self,
        symbol: Symbol,
        value: float,
        asap: bool,
        price: Optional[float]
    ) -> None:
        self.symbol = symbol
        self.value = value

        # If price is fixed, then we could not trade asap,
        # because we could not set the price for market limit order
        self.asap = False if price is not None else asap
        self.price = price

    def __repr__(self) -> str:
        return class_repr(self, main='symbol')

    def equals_to(
        self,
        position: 'SymbolPosition'
    ) -> bool:
        """To detect whether the given `SymbolPosition` has the same goal of the current one
        """

        return (
            self.asap == position.asap
            and self.value == position.value
            and self.price == position.price
        )


class SymbolInfo:
    __slots__ = (
        'symbol',

        'min_price',
        'max_price',

        # Minimum price step of a symbol
        'min_price_step',
        '_min_price_step_precision',

        'min_quantity',
        'max_quantity',

        # Minimum quantity step of an asset based on quote asset
        'min_quantity_step',
        'min_quantity_step_precision',

        # Minimum value which is the product of price and quantity
        'min_notional'
    )

    symbol: Symbol

    min_price: float
    max_price: float
    min_price_step: float

    min_quantity: float
    max_quantity: float
    min_quantity_step: float
    min_quantity_step_precision: int

    min_notional: float

    def __init__(
        self,
        symbol: str,
        base_asset: str,
        quote_asset: str,

        min_price: str,
        max_price: str,
        min_price_step: str,

        min_quantity: str,
        max_quantity: str,
        min_quantity_step: str,

        # An order's notional value is the `price` * `quantity`
        min_notional: str
    ):
        self.symbol = Symbol(
            symbol,
            base_asset,
            quote_asset
        )

        self.min_price = float(min_price)
        self.max_price = float(max_price)

        self.min_price_step = float(min_price_step)
        self._min_price_step_precision = len(
            float_to_str(self.min_price_step)
        ) - 2

        self.min_quantity = float(min_quantity)
        self.max_quantity = float(max_quantity)

        self.min_quantity_step = float(min_quantity_step)
        self.min_quantity_step_precision = len(
            float_to_str(self.min_quantity_step)
        ) - 2

        self.min_notional = float(min_notional)

    def __repr__(self) -> str:
        price = format(
            self.min_price_step,
            f'.{self._min_price_step_precision}f'
        )

        quantity = format(
            self.min_quantity_step,
            f'.{self.min_quantity_step_precision}f'
        )

        return f'<SymbolInfo {self.symbol}: price_step >= {price}, quantity_step >= {quantity}>'


class TicketOrderSide(Enum):
    def __str__(self):
        return self.value

    # USDT is locked
    BUY = 'BUY'
    # BTC is locked
    SELL = 'SELL'


class TicketOrderStatus(Enum):
    def __str__(self):
        return self.value[0]

    # The ticket is initialized but has not been submitted to the exchange,
    # or the ticket is failed to create order so back to the initial state
    INIT = 'INIT'

    # The ticket order is creating,
    #   the request is about to send to the exchange,
    #   but not yet get response
    SUBMITTING = 'SUBMITTING'

    # The ticket order is created via the exchange API
    CREATED = 'CREATED'

    def lt(self, status: 'TicketOrderStatus') -> bool:
        """
        Returns `True` if the current status is less than the given status
        """

        return self.value[1] <= status.value[1]


class OrderTicket:
    __slots__ = (
        'id',
        'order_side',
        'symbol',
        'price',
        'quantity',
        'locked_asset',
        'locked_quantity',
        'position',
        'info',
        'status',
        'filled_quantity',
        'time'
    )

    _UID: int = 0

    id: int
    order_side: TicketOrderSide
    symbol: Symbol
    price: float
    quantity: str
    locked_asset: str
    locked_quantity: float
    position: SymbolPosition
    info: SymbolInfo
    status: TicketOrderStatus
    filled_quantity: str
    time: str

    def __init__(
        self,

        # Order type, which will be used by trader
        order_side: TicketOrderSide,
        # The related symbol
        symbol: Symbol,
        # price of symbl
        price: float,
        # We should use str as quantity
        quantity: str,

        # The quantity of which asset has been locked,
        # could be either target asset and cash asset
        locked_asset: str,
        # locked quantity
        locked_quantity: float,

        position: SymbolPosition,
        info: SymbolInfo
    ) -> None:
        self.id = OrderTicket._UID

        OrderTicket._UID += 1

        self.order_side = order_side

        self.symbol = symbol
        self.price = price
        self.quantity = quantity

        self.locked_asset = locked_asset
        self.locked_quantity = locked_quantity

        self.position = position
        self.info = info

        self.status = TicketOrderStatus.INIT

        self.filled_quantity = '0'

        # self.time = datetime_now_str()

    __repr__ = class_repr


class Balance:
    __slots__ = (
        'asset',
        'free',
        'locked'
    )

    asset: str
    free: float
    locked: float

    def __init__(
        self,
        asset: str,
        free: float,
        locked: float
    ):
        self.asset = asset
        self.free = free
        self.locked = locked

    def __repr__(self) -> str:
        return class_repr(self, main='asset')

    def exists(self) -> bool:
        """
        If the total balance is 0, then the asset does not exist
        """

        return self.free + self.locked != 0.


class TicketGroup:
    """
    Group sell and group buy are mutually exclusive
    """

    __slots__ = (
        'buy',
        'sell'
    )

    # The tickets that buy the asset
    buy: Set[OrderTicket]
    # The tickets that sell the asset
    sell: Set[OrderTicket]

    def __init__(self):
        self.buy = set()
        self.sell = set()

    def get(
        self,
        direction: TicketOrderSide
    ) -> Set[OrderTicket]:
        """
        Get a copied group of the given direction
        """

        return (
            self.buy if direction is TicketOrderSide.BUY else self.sell
        ).copy()

    def close(self, ticket: OrderTicket) -> None:
        """
        Close a ticket
        """

        (self.buy or self.sell).discard(ticket)

    def clear(self) -> None:
        """
        Clear all tickets
        """

        (self.buy or self.sell).clear()


FuncCancelOrder = Callable[[OrderTicket], Awaitable[None]]



import math
import decimal

from typing import (
    Optional,
    Tuple
)


def apply_precision(value: float, decimals: int) -> str:
    """
    Stringify a float according to the given precision without rounding

    Usage::

        apply_precision(0.123456, 2)  # '0.12'
    """

    magnitude = 10 ** decimals

    return format(
        # formatting is always with rounding,
        # so, we must
        math.floor(value * magnitude) / magnitude,
        f'.{decimals}f'
    )


def class_repr(
    self,
    main: Optional[str] = None,
    keys: Optional[Tuple[str]] = None
) -> str:
    """
    """

    Class = type(self)

    slots = Class.__slots__ if keys is None else keys

    string = f'<{Class.__name__}'

    if main is not None:
        string += f' {getattr(self, main)}'

    string += ': '

    string += ', '.join([
        f'{name}: {getattr(self, name)}'
        for name in slots if name != main
    ])

    string += '>'

    return string


decimal_ctx = decimal.Context()
decimal_ctx.prec = 20


def float_to_str(f: float) -> str:
    """
    Convert the given float to a string,
    without resorting to scientific notation
    """

    d = decimal_ctx.create_decimal(repr(f))
    return format(d, 'f')



from trading_state.types import (
    SymbolInfo,
    Balance
)


def test_float_precision():
    info = SymbolInfo(
        'BTCUSDT',
        'BTC',
        'USDT',

        '0.010000000000',
        '1000000',
        '0.010000000000',

        '0.000001000000',
        '9000',
        '0.000001000000',

        '10.00000000'
    )

    assert info.min_quantity_step_precision == 6


def test_balance():
    balance = Balance('BTCUSDT', 0, 0)

    assert not balance.exists()



from .types import Symbol

from .util import float_to_str


# SYMBOL_FILTER_TYPE_LOT_SIZE = 'lot_size'


class SymbolInfo:
    __slots__ = (
        'symbol',

        'min_price',
        'max_price',

        # Minimum price step of a symbol
        'min_price_step',
        '_min_price_step_precision',

        'min_quantity',
        'max_quantity',

        # Minimum quantity step of an asset based on quote asset
        'min_quantity_step',
        'min_quantity_step_precision',

        # Minimum value which is the product of price and quantity
        'min_notional'
    )

    symbol: Symbol

    min_price: float
    max_price: float
    min_price_step: float

    min_quantity: float
    max_quantity: float
    min_quantity_step: float
    min_quantity_step_precision: int

    min_notional: float

    def __init__(
        self,
        symbol: str,
        base_asset: str,
        quote_asset: str,

        min_price: str,
        max_price: str,
        min_price_step: str,

        min_quantity: str,
        max_quantity: str,
        min_quantity_step: str,

        # An order's notional value is the `price` * `quantity`
        min_notional: str
    ):
        self.symbol = Symbol(
            symbol,
            base_asset,
            quote_asset
        )

        self.min_price = float(min_price)
        self.max_price = float(max_price)

        self.min_price_step = float(min_price_step)
        self._min_price_step_precision = len(
            float_to_str(self.min_price_step)
        ) - 2

        self.min_quantity = float(min_quantity)
        self.max_quantity = float(max_quantity)

        self.min_quantity_step = float(min_quantity_step)
        self.min_quantity_step_precision = len(
            float_to_str(self.min_quantity_step)
        ) - 2

        self.min_notional = float(min_notional)

    def add_filter () -> None:


    def __repr__(self) -> str:
        price = format(
            self.min_price_step,
            f'.{self._min_price_step_precision}f'
        )

        quantity = format(
            self.min_quantity_step,
            f'.{self.min_quantity_step_precision}f'
        )

        return f'<SymbolInfo {self.symbol}: price_step >= {price}, quantity_step >= {quantity}>'



# import asyncio
import pytest

from trading_state.state import (
    TraderState,

    # TicketOrderSide
)
from trading_state.types import (
    Balance
)

BTC = 'BTC'
USDT = 'USDT'
SYMBOL = BTC + USDT


def set_price(state: TraderState):
    state.set_price(SYMBOL, 7000.)


def set_symbol_info(state: TraderState):
    state.set_symbol_info(
        SYMBOL,
        BTC,
        USDT,

        '0.01',
        '1000000',
        '0.01',

        '0.000001',
        '9000',
        '0.000001',

        '10.00000000'
    )


def set_balances(state: TraderState):
    state.update_balances([
        Balance(
            BTC,
            15.,
            0.
        ),
        Balance(
            USDT,
            0.001,
            0.
        )
    ])


def set_info(state: TraderState):
    set_price(state)
    set_symbol_info(state)


def test_support():
    state = TraderState()

    set_info(state)

    assert state.support_symbol(SYMBOL)

    LTEBTC = 'LTEBTC'

    assert not state.support_symbol(LTEBTC)
    assert state.create_ticket(
        LTEBTC,
        1.,
        False,
        None
    ) == (None, set())


def test_non_diff():
    state = TraderState()

    set_balances(state)
    set_info(state)

    state.expect(
        SYMBOL,
        1.,
        False,
        None
    )

    diffed, to_cancel = state.get_tickets()

    assert not diffed
    assert not to_cancel


def test_specified_price():
    state = TraderState()

    set_balances(state)
    set_info(state)

    ticket, to_cancel = state.create_ticket(
        SYMBOL,
        0.,
        False,
        7000.
    )

    assert not to_cancel
    assert ticket is not None
    assert ticket.price == 7000.

    ticket2, to_cancel2 = state.create_ticket(
        SYMBOL,
        0.,
        False,
        6000.
    )

    assert ticket in to_cancel2
    assert ticket2.price == 6000.

    state.expect(
        SYMBOL,
        1.,
        False,
        5000.
    )

    # Different direction, cancel previous ticket
    # ---------------------------------------------

    tickets, to_cancel = state.get_tickets()

    assert ticket2 in to_cancel
    assert not tickets

    # get_tickets again, previous ticket already closed
    # ---------------------------------------------

    tickets, to_cancel = state.get_tickets()

    assert not tickets
    assert not tickets

    # Update balances should reset diff

    state.update_balances([
        Balance(
            USDT,
            1000.,
            0.
        )
    ])

    tickets, to_cancel = state.get_tickets()

    ticket = tickets[0]

    assert ticket.price == 5000.
    assert ticket.symbol.name == SYMBOL
    assert not to_cancel

    # Will remove expectations

    state.create_ticket(
        SYMBOL,
        0.,
        False,
        None
    )

    assert not state._expected


@pytest.mark.asyncio
async def test_diff():
    state = TraderState()

    def expect_sell_all(quantity: float = 15.):
        # sell all
        state.expect(
            SYMBOL,
            0.,
            False,
            None
        )

        if not state._expected:
            return

        assert repr(state._expected[SYMBOL]) == '<SymbolPosition BTCUSDT: value: 0.0, asap: False, price: None>'

        diff, _ = state.get_tickets()

        if not diff:
            return

        ticket = diff[0]

        assert str(ticket.order_side) == 'SELL'
        assert ticket.symbol.name == SYMBOL
        assert ticket.price == 7000.

        assert ticket.quantity == '15.000000'

        assert ticket.locked_asset == BTC
        assert ticket.locked_quantity == quantity

        return ticket

    def expect_all_in():
        # buy all
        state.expect(
            SYMBOL,
            1.,
            False,
            None
        )

        diff, _ = state.get_tickets()

        if not diff:
            return

        return diff[0]

    state.expect(
        SYMBOL,
        0.,
        False,
        None
    )

    assert not state._expected, 'BTC not supported'

    tickets, to_cancel = state.get_tickets()
    assert not tickets
    assert not to_cancel

    assert expect_sell_all() is None, 'symbol info is not set'

    set_symbol_info(state)

    assert repr(state._symbol_infos[SYMBOL]) == '<SymbolInfo BTCUSDT: price_step >= 0.01, quantity_step >= 0.000001>'

    assert expect_sell_all() is None, 'price is not set'

    state.set_price(SYMBOL, 7000.)

    assert expect_sell_all() is None, 'balances are not set'

    state.update_balances([
        Balance(
            BTC,
            15.,
            0.
        ),
        Balance(
            USDT,
            0.001,
            0.
        )
    ])

    assert repr(state._balances[BTC]) == '<Balance BTC: free: 15.0, locked: 0.0>'

    ticket0 = expect_sell_all()
    assert ticket0 is not None

    # Unsolved tickets will be always returned by state.tickets

    tickets, _ = state.get_tickets()
    assert tickets[0] is ticket0

    # ticket 0 will be closed
    assert expect_all_in() is None

    ticket1 = expect_sell_all()

    assert ticket1 is not None

    # Test to close ticket duplicately
    # ------------------------------------------------------

    state.close_ticket(ticket1)
    state.close_ticket(ticket1)

    # Test to create ticket
    # ------------------------------------------------------

    ticket_sell, to_cancel = state.create_ticket(
        SYMBOL,
        0.,
        False,
        None
    )

    assert ticket_sell.position.value == 0.
    assert ticket_sell.symbol.name == SYMBOL
    assert not to_cancel

    # There is already a ticket
    # ------------------------------------------------------

    ticket_sell2, to_cancel2 = state.create_ticket(
        SYMBOL,
        0.,
        False,
        None
    )

    assert ticket_sell2 is None
    assert not to_cancel2

    ticket_buy, to_cancel = state.create_ticket(
        SYMBOL,
        1.,
        False,
        None
    )

    assert ticket_buy is None
    assert ticket_sell in to_cancel

    assert not state._expected

    state.update_balances([
        Balance(
            BTC,
            0.,
            0.
        ),
        Balance(
            USDT,
            70000,
            0.
        )
    ])

    assert expect_all_in() is not None
    assert expect_sell_all() is None

    # Set quota to 0
    state.set_quota(BTC, 0.)
    state.set_quota('ETH', 0.)

    assert expect_all_in() is None, 'not enough quota'

    state.set_quota(BTC, None)

    state.update_balances([
        Balance(
            USDT,
            0.006,
            0.
        )
    ])

    assert expect_all_in() is None, 'min price step'

    state.update_balances([
        Balance(
            USDT,
            9.,
            0.
        )
    ])

    assert expect_all_in() is None, 'min notional'

    state.update_balances([
        Balance(
            USDT,
            15.,
            0.
        )
    ])

    state.set_price(SYMBOL, 70000000.)
    state.set_price(SYMBOL, 70000000.)

    assert expect_all_in() is None, 'not enough'

    state.set_price(SYMBOL, 7000.)

    ticket2 = expect_all_in()

    assert ticket2 is not None

    assert ticket2.quantity == '0.002142'

    state.clear()

    tickets, to_cancel = state.get_tickets()

    assert not tickets
    assert not to_cancel

    assert not state._expected
